@@ -67,6 +67,7 @@ static NvBool DpySetBacklightBrightness(NVDpyEvoRec *pDpyEvo, NvS64 brightness)
     params.subDeviceInstance = pDispEvo->displayOwner;
     params.displayId = nvDpyEvoGetConnectorId(pDpyEvo);
     params.brightness = brightness;
+    params.brightnessType = NV0073_CTRL_SPECIFIC_BACKLIGHT_BRIGHTNESS_TYPE_PERCENT100;
 
     ret = nvRmApiControl(
             nvEvoGlobal.clientHandle,
@@ -97,6 +98,7 @@ static NvBool DpyGetBacklightBrightness(const NVDpyEvoRec *pDpyEvo,
 
     params.subDeviceInstance = pDispEvo->displayOwner;
     params.displayId = nvDpyEvoGetConnectorId(pDpyEvo);
+    params.brightnessType = NV0073_CTRL_SPECIFIC_BACKLIGHT_BRIGHTNESS_TYPE_PERCENT100;
 
     ret = nvRmApiControl(
             nvEvoGlobal.clientHandle,
@@ -230,7 +232,7 @@ static void SetDitheringCommon(NVDpyEvoPtr pDpyEvo)
              (nvDpyIdIsInDpyIdList(pDpyEvo->id, pApiHeadState->activeDpys)));
 
     nvChooseDitheringEvo(pConnectorEvo,
-                         pApiHeadState->attributes.colorBpc,
+                         pApiHeadState->attributes.color.bpc,
                          &pDpyEvo->requestedDithering,
                          &pApiHeadState->attributes.dithering);
 
@@ -608,11 +610,7 @@ static void DpyPostColorSpaceOrRangeSetEvo(NVDpyEvoPtr pDpyEvo)
     NVEvoUpdateState updateState = { };
     NVDispEvoRec *pDispEvo = pDpyEvo->pDispEvo;
     NVDispApiHeadStateEvoRec *pApiHeadState;
-    NvU8 hdmiFrlBpc;
     NvU32 head;
-#if defined(DEBUG)
-    NvU32 hwHead;
-#endif
 
     if (pDpyEvo->apiHead == NV_INVALID_HEAD) {
         return;
@@ -622,23 +620,13 @@ static void DpyPostColorSpaceOrRangeSetEvo(NVDpyEvoPtr pDpyEvo)
     nvAssert((pApiHeadState->hwHeadsMask) != 0x0 &&
              (nvDpyIdIsInDpyIdList(pDpyEvo->id, pApiHeadState->activeDpys)));
 
-    head = nvGetPrimaryHwHead(pDispEvo, pDpyEvo->apiHead);
-    hdmiFrlBpc = pDispEvo->headState[head].hdmiFrlBpc;
-#if defined(DEBUG)
-    FOR_EACH_EVO_HW_HEAD_IN_MASK(pApiHeadState->hwHeadsMask, hwHead) {
-        nvAssert(pDispEvo->headState[head].timings.yuv420Mode ==
-                 pDispEvo->headState[hwHead].timings.yuv420Mode);
-    }
-#endif
-
     /*
      * Choose current colorSpace and colorRange based on the current mode
      * timings and the requested color space and range.
      */
     if (!nvChooseCurrentColorSpaceAndRangeEvo(pDpyEvo,
-                                              &pDispEvo->headState[head].timings,
-                                              hdmiFrlBpc,
-                                              pApiHeadState->colorimetry,
+                                              pApiHeadState->timings.yuv420Mode,
+                                              pApiHeadState->attributes.color.colorimetry,
                                               pDpyEvo->requestedColorSpace,
                                               pDpyEvo->requestedColorRange,
                                               &colorSpace,
@@ -650,26 +638,68 @@ static void DpyPostColorSpaceOrRangeSetEvo(NVDpyEvoPtr pDpyEvo)
 
     /* For DP, neither color space nor bpc can be changed without a modeset */
     if (nvConnectorUsesDPLib(pDpyEvo->pConnectorEvo) &&
-            ((pApiHeadState->attributes.colorSpace != colorSpace) ||
-             (pApiHeadState->attributes.colorBpc != colorBpc))) {
+            ((pApiHeadState->attributes.color.format != colorSpace) ||
+             (pApiHeadState->attributes.color.bpc != colorBpc))) {
         return;
     }
 
-    pApiHeadState->attributes.colorSpace = colorSpace;
-    pApiHeadState->attributes.colorRange = colorRange;
-    pApiHeadState->attributes.colorBpc = colorBpc;
+    /*
+     * Hardware does not support HDMI FRL with YUV422, and it is not possible
+     * to downgrade current color bpc on HDMI FRL at this point.
+     */
+    if ((pApiHeadState->timings.protocol == NVKMS_PROTOCOL_SOR_HDMI_FRL) &&
+            ((colorSpace == NV_KMS_DPY_ATTRIBUTE_CURRENT_COLOR_SPACE_YCbCr422) ||
+             (pApiHeadState->attributes.color.bpc > colorBpc))) {
+        return;
+    }
+
+    if (nvDpyIsHdmiEvo(pDpyEvo) &&
+            (colorBpc > pApiHeadState->attributes.color.bpc)) {
+        NVDpyAttributeColor tmpDpyColor = pApiHeadState->attributes.color;
+
+        tmpDpyColor.format = colorSpace;
+        tmpDpyColor.range = colorRange;
+        tmpDpyColor.bpc = colorBpc;
+
+        /*
+         * For HDMI FRL, downgrade the selected color bpc to the current color
+         * bpc so that the current color bpc remains unchanged.
+         */
+        if (pApiHeadState->timings.protocol == NVKMS_PROTOCOL_SOR_HDMI_FRL) {
+            tmpDpyColor.bpc = pApiHeadState->attributes.color.bpc;
+        } else {
+            const NVColorFormatInfoRec colorFormatsInfo =
+                nvGetColorFormatInfo(pDpyEvo);
+
+            while (nvHdmiGetEffectivePixelClockKHz(pDpyEvo,
+                                                   &pApiHeadState->timings,
+                                                   &tmpDpyColor) >
+                       pDpyEvo->maxSingleLinkPixelClockKHz) {
+
+                if(!nvDowngradeColorSpaceAndBpc(&colorFormatsInfo,
+                                                &tmpDpyColor)) {
+                    return;
+                }
+            }
+        }
+
+        pApiHeadState->attributes.color.format = tmpDpyColor.format;
+        pApiHeadState->attributes.color.range = tmpDpyColor.range;
+        pApiHeadState->attributes.color.bpc = tmpDpyColor.bpc;
+    } else {
+        pApiHeadState->attributes.color.format = colorSpace;
+        pApiHeadState->attributes.color.range = colorRange;
+        pApiHeadState->attributes.color.bpc = colorBpc;
+    }
 
     /* Update hardware's current colorSpace and colorRange */
     FOR_EACH_EVO_HW_HEAD_IN_MASK(pApiHeadState->hwHeadsMask, head) {
         enum nvKmsPixelDepth newPixelDepth =
-            nvEvoColorSpaceBpcToPixelDepth(pApiHeadState->attributes.colorSpace,
-                                           pApiHeadState->attributes.colorBpc);
+            nvEvoDpyColorToPixelDepth(&pApiHeadState->attributes.color);
 
         nvUpdateCurrentHardwareColorSpaceAndRangeEvo(pDispEvo,
                                                      head,
-                                                     pApiHeadState->colorimetry,
-                                                     pApiHeadState->attributes.colorSpace,
-                                                     pApiHeadState->attributes.colorRange,
+                                                     &pApiHeadState->attributes.color,
                                                      &updateState);
 
         if (newPixelDepth != pDispEvo->headState[head].pixelDepth) {
@@ -737,7 +767,7 @@ static NvBool GetCurrentColorSpace(const NVDpyEvoRec *pDpyEvo, NvS64 *pValue)
         return FALSE;
     }
 
-    *pValue = pDpyEvo->currentAttributes.colorSpace;
+    *pValue = pDpyEvo->currentAttributes.color.format;
 
     return TRUE;
 }
@@ -810,7 +840,7 @@ static NvBool GetCurrentColorRange(const NVDpyEvoRec *pDpyEvo, NvS64 *pValue)
         return FALSE;
     }
 
-    *pValue = pDpyEvo->currentAttributes.colorRange;
+    *pValue = pDpyEvo->currentAttributes.color.range;
 
     return TRUE;
 }
